"""This module contains the managers for the ``django-discover`` app"""

import io
import os
from zipfile import ZipFile, ZIP_LZMA

from django.conf import settings
from django.core.files.base import ContentFile
from django.core.files.storage import default_storage
from django.core.serializers import deserialize, serialize
from django.db import transaction
from django.db.models import (
    Manager,
    Q,
    QuerySet
)

import requests
from psycopg2.extras import NumericRange


class DiscoverItemQueryset(QuerySet):
    """QuerySet for the ``DiscoverItem`` model"""

    def for_language(self, language):
        """Filter the discover items for the given language

        Parameters
        ----------
        language: str
            The language to filter on

        Returns
        -------
        DiscoverItemQueryset
            The filtered Queryset

        """

        return self.filter(language=language)

    def for_version(self, major_version, minor_version=0, hotfix_version=0):
        """Filter the discover items for the given version

        The query below can be decomposed like this:
        - Return all the item for which the major versions field is not filled
        - Return the items for which the major version field is filled IF:
          - The given version is in the bounds
          - AND The minor version is in the bounds (if filled)
          - AND The hotfix version is in the bounds (if filled)

        Parameters
        ----------
        major_version: int
            The major version
        minor_version: int
            The minor version
        hotfix_version: int
            The hotfix version

        Returns
        -------
        DiscoverItemQueryset
            The filtered Queryset

        """

        return self.filter(
            # The major version was not filled
            Q(major_versions__isnull=True) |
            Q(
                # The major versions contains the current version
                Q(
                    major_versions__contains=NumericRange(
                        major_version,
                        major_version,
                        bounds='[]'
                    )
                ) &
                Q(
                    Q(minor_versions__isnull=True) |
                    Q(
                        minor_versions__isnull=False,
                        minor_versions__contains=NumericRange(
                            minor_version,
                            minor_version,
                            bounds='[]'
                        )
                    )
                ) &
                Q(
                    Q(hotfix_versions__isnull=True) |
                    Q(
                        hotfix_versions__isnull=False,
                        hotfix_versions__contains=NumericRange(
                            hotfix_version,
                            hotfix_version,
                            bounds='[]'
                        )
                    )
                )
            )
        )


class DiscoverItemManager(Manager.from_queryset(DiscoverItemQueryset)):
    """Manager for the ``DiscoverItem`` model"""

    @staticmethod
    def get_zipped_discover_items(discover_items_queryset):
        """Get a ZipFile containing the serialized discover items queryset

        Parameters
        ----------
        discover_items_queryset: DiscoverItemQueryset
            The filtered queryset containing the discover items to serialize

        Returns
        -------
        io.BytesIO
            A buffer of a ZipFile containing the following elements:
            - A json file containing the serialized discover items
            - The media of the serialized items

        """

        # Build the ZipFile
        zip_buffer = io.BytesIO()
        zip_file = ZipFile(zip_buffer, 'w', ZIP_LZMA)

        # Copy the serialized discover items
        serialized_items = serialize('json', discover_items_queryset)
        zip_file.writestr('discover_items.json', serialized_items)

        # Copy the media
        medias_list = discover_items_queryset.values_list('image', flat=True)
        for media in medias_list:
            file_path = os.path.join(settings.MEDIA_ROOT, media)
            zip_file.write(file_path, media)

        zip_file.close()

        return zip_buffer

    @staticmethod
    def _load_zip_archive(from_url=None, from_storage_path=None, from_local_path=None):
        """Load an archive containing serialized discover items

        For the parameters, see ``load_zipped_discover_items``

        Parameters
        ----------
        from_url : str
            If given, the zip will be downloaded from this url.
        from_storage_path : str
            If given, the zip will be opened from this path from the django default storage.
            Can be an absolute or relative path. If absolute, it must match the default
            storage root. If relative, it must be relative to the default storage root.
            (storage root is generally ``settings.MEDIA_ROOT``.
        from_local_path : str
            If given, the zip will be opened from this path from the local filesystem.
            Must be an absolute path.

        Returns
        -------
        Optional[ZipFile]
            The loaded zipfile, in memory.
            If not data from url (url not found) or path (path not exists in storage or local disk)
            then returns ``None``.

        Raises
        ------
        AssertionError
            If none or many of ``from_url`` and ``from_storage_path``, ``from_local_path`` are given

        """

        assert bool(from_url) ^ bool(from_storage_path) ^ bool(from_local_path), \
            "Expect one and only one of `from_url`, `from_storage_path` and `from_local_path`"

        if from_url:
            response = requests.get(from_url, stream=True)
            try:
                response.raise_for_status()
            except requests.HTTPError:
                # A 404 error means no discover items were available
                if response.status_code == 404:
                    return None
                # If we encounter another error, re-raise the ``HTTPError``
                raise

            zip_data = response.content

        elif from_storage_path:
            if not default_storage.exists(from_storage_path):
                # No file means no discover items were available
                return None

            with default_storage.open(from_storage_path, 'rb') as file_d:
                zip_data = file_d.read()

        elif from_local_path:
            if not os.path.exists(from_local_path):
                # No file means no discover items were available
                return None

            with open(from_local_path, 'rb') as file_d:
                zip_data = file_d.read()

        return ZipFile(io.BytesIO(zip_data), 'r', ZIP_LZMA)

    def _load_zipped_discover_items(self, zip_file):
        """Load an archive containing serialized discover items

        Parameters
        ----------
        zip_file : ZipFile
            The zip, loaded in memory, with data to extract.

        """

        with transaction.atomic():

            # 1 - Erase all of the actual discover items
            self.get_queryset().all().delete()

            # 2 - Load the new discover items objects from the json file
            images = {}
            for deserialized in deserialize('json', zip_file.read('discover_items.json')):
                deserialized.save()
                if deserialized.object.image:
                    images[deserialized.object.image.name] = deserialized.object.image.path

        # 3 - Now that data is saved in db, we can save images
        for image_name, image_path in images.items():
            try:
                default_storage.save(
                    name=image_path,
                    content=ContentFile(zip_file.read(image_name))
                )
            except Exception:  # pylint: disable=broad-except
                # file is not in the zip, we don't abort and try to save the other images
                pass

    def load_zipped_discover_items(self,
                                   from_url=None, from_storage_path=None, from_local_path=None):
        """Load an archive containing serialized discover items

        Parameters
        ----------
        from_url : str
            If given, the zip will be downloaded from this url.
        from_storage_path : str
            If given, the zip will be opened from this path from the django default storage.
            Must be an absolute path.
        from_local_path : str
            If given, the zip will be opened from this path from the local filesystem.
            Must be an absolute path.

        Returns
        -------
        bool
            ``True`` if an archive was found and imported. ``False`` if no archive was found.

        Raises
        ------
        ValueError
            If any error occurred while getting or extracting the archive, or while
            saving new items from this archive.

        """

        try:
            zip_file = self._load_zip_archive(from_url, from_storage_path, from_local_path)
            if zip_file is None:
                return False

            self._load_zipped_discover_items(zip_file)
            return True

        except Exception as exception:  # pylint: disable=broad-except

            # We raise a specific error
            raise ValueError(
                'Error during the discover items fetching. The exception was "%s"' % exception
            )




# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2018-01-26 00:53
from __future__ import unicode_literals

from django.db import migrations, models
import django_discover.fields


class Migration(migrations.Migration):

    dependencies = [
        ('django_discover', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='discoveritem',
            name='is_active',
            field=models.BooleanField(default=True, verbose_name='Is active'),
        ),
    ]



"""This module contains the ``DiscoverItem``, to expose a functionality to the user"""

from django.contrib.postgres.fields import IntegerRangeField
from django.db import models
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _

from extended_choices.choices import Choices

from .fields import LanguageField
from .managers import DiscoverItemManager


class DiscoverItem(models.Model):
    """Model to display an introduction to a given feature

    Attributes
    ----------
    TAG_CHOICES: ``Choices``
        A list of available tags
    CATEGORY_CHOICES: ``Choices``
        A list of available categories

    language: django ``LanguageField``
        The language of the current discover item

    is_active: django ``BooleanField``
        Is the  current discover item is active or not ?

    title: django ``CharField``
        The title of the current discover item
    slug: django ``SlugField``
        The slug of the current discover item
    image: django ``ImageField``
        The image of the current discover item
    short_description: django ``CharField``
        A short description of the current discover item
    long_description: django ``TextField``
        The complete description of the current discover item

    link_title: django ``CharField``
        The title of the link of this discover item
    link_url: django ``CharField``
        The url of the link of this discover item

    tag: django ``CharField``
        A tag for the current discover item, contains one of the ``TAG_CHOICES``
    category: django ``CharField``
        A category for the current discover item, contains one of the ``CATEGORY_CHOICES``

    major_versions: django ``IntegerRangeField``
        The major versions range for this discover item to be displayed
    minor_versions: django ``IntegerRangeField``
        The minor versions range for this discover item to be displayed
    hotfix_versions: django ``IntegerRangeField``
        The hotfix versions range for this discover item to be displayed

    created_at: django ``DateTimeField``
        The date and time of the creation of this discover item
    updated_at: django ``DateTimeField``
        The date and time of the last update of this discover item

    """

    TAG_CHOICES = Choices(
        ('NEW', 'new', _('New')),
        ('IMPORTANT', 'important', _('Important')),
    )

    CATEGORY_CHOICES = Choices(
        ('RESOURCES', 'resources', _('Resources')),
        ('SESSIONS', 'sessions', _('Sessions')),
        ('USERS', 'users', _('Users')),
    )

    is_active = models.BooleanField(
        default=True,
        verbose_name=_('Is active'),
    )

    language = LanguageField(
        verbose_name=_('Language'),
        help_text=_('The language of this discover item')
    )

    title = models.CharField(
        max_length=30,
        verbose_name=_('Title')
    )
    slug = models.SlugField(
        max_length=500,
        verbose_name=_('Slug')
    )
    image = models.ImageField(
        verbose_name=_('Image'),
        upload_to='django_discover'
    )
    short_description = models.CharField(
        max_length=80,
        verbose_name=_('Short description')
    )
    long_description = models.TextField(
        verbose_name=_('Long description')
    )

    link_title = models.CharField(
        max_length=50,
        verbose_name=_('Link title'),
        null=True,
        blank=True
    )
    link_url = models.CharField(
        max_length=500,
        verbose_name=_('Link url'),
        null=True,
        blank=True
    )

    tag = models.CharField(
        max_length=20,
        choices=TAG_CHOICES,
        verbose_name=_('Tag'),
        null=True,
        blank=True
    )
    category = models.CharField(
        max_length=20,
        choices=CATEGORY_CHOICES,
        verbose_name=_('Category'),
        null=True,
        blank=True
    )

    major_versions = IntegerRangeField(
        verbose_name=_('Major version range'),
        null=True,
        blank=True
    )
    minor_versions = IntegerRangeField(
        verbose_name=_('Minor version range'),
        null=True,
        blank=True
    )
    hotfix_versions = IntegerRangeField(
        verbose_name=_('Hotfix version range'),
        null=True,
        blank=True
    )

    created_at = models.DateTimeField(
        auto_now_add=True,
        verbose_name=_('Date and time of creation'),
    )
    updated_at = models.DateTimeField(
        auto_now=True,
        verbose_name=_('Date and time of the last update'),
    )

    objects = DiscoverItemManager()

    class Meta:
        unique_together = (('language', 'slug'),)

    def __str__(self):
        return '[%(language)s] %(slug)s <%(from_version)s - %(until_version)s>' % {
            'language': self.language,
            'slug': self.slug,
            'from_version': self.from_version,
            'until_version': self.until_version
        }

    @cached_property
    def from_version(self):
        """Get the full version number from this item should be displayed

        Returns
        -------
        str

        """

        def version_lower_part_to_string(version):
            """Make the lower bound of the version part to a string

            Parameters
            ----------
            version: IntegerRangeField

            Returns
            -------
            str
                The lower version of the given part
                or '*'

            """

            if version and version.lower:
                return version.lower

            return '*'

        return '%s.%s.%s' % (
            version_lower_part_to_string(self.major_versions),
            version_lower_part_to_string(self.minor_versions),
            version_lower_part_to_string(self.hotfix_versions)
        )

    @cached_property
    def until_version(self):
        """Get the full version number until this item should not be displayed

        Returns
        -------
        str

        """

        def version_upper_part_to_string(version):
            """Make the upper bound of the version part to a string

            Parameters
            ----------
            version: IntegerRangeField

            Returns
            -------
            str
                The upper version of the given part
                or '*'

            """

            if version and version.upper:
                return version.upper

            return '*'

        return '%s.%s.%s' % (
            version_upper_part_to_string(self.major_versions),
            version_upper_part_to_string(self.minor_versions),
            version_upper_part_to_string(self.hotfix_versions),
        )



"""
``django-discover`` provides ``DiscoverItem`` model, that allow displaying an introduction
to an existent or a new feature to the user.
"""

__author__ = 'Clément "clemfromspace" Denoix'
__contact__ = 'clement.denoix@magency.fr'
__homepage__ = 'https://gitlab.com/magency/'
__license__ = 'BSD'
__version__ = '1.0.1'
VERSION = tuple(__version__.split('.'))




"""This module contains the ``DiscoverItem`` admin class"""

from django.contrib import admin

from .models import DiscoverItem


@admin.register(DiscoverItem)
class DiscoverItemAdmin(admin.ModelAdmin):
    """Admin model for the ``DiscoverItem`` model"""

    prepopulated_fields = {'slug': ('title',)}



"""This module contains the factories for the ``django_discover`` app"""

import factory
from django.conf import settings

from .models import DiscoverItem


class DiscoverItemFactory(factory.django.DjangoModelFactory):
    """Factory for the ``DiscoverItem`` model"""

    language = settings.LANGUAGES[0][0]
    title = factory.Sequence(lambda n: 'discover item %d' % n)
    slug = factory.Sequence(lambda n: 'discover-item-%d' % n)
    image = factory.django.ImageField()
    short_description = factory.Sequence(lambda n: 'short description discover item %d' % n)
    long_description = factory.Sequence(lambda n: 'long description discover item %d' % n)

    class Meta:
        model = DiscoverItem



"""This module contains the urls for the ``django_discover`` app"""

from django.conf.urls import url

from .views import DiscoverItemsArchiveView

app_name = 'django_discover'  # pylint: disable=invalid-name

urlpatterns = [
    url(
        r'^discover_items/archive/(?P<version>[0-9]+.[0-9]+.[0-9]+)/$',
        DiscoverItemsArchiveView.as_view(),
        name='discover-items-archive'
    ),
]
